@@ -121,7 +121,7 @@ paCallback(void *inputBuffer, void *outputBuffer,
 
 	driver->inPortAudio = (float*)inputBuffer;
 	driver->outPortAudio = (float*)outputBuffer;
-	driver->last_wait_ust = jack_get_microseconds();
+	driver->last_wait_ust = driver->engine->get_microseconds();
 	return driver->engine->run_cycle(driver->engine, framesPerBuffer, 0);
 }
 
@@ -132,10 +132,14 @@ portaudio_driver_attach (portaudio_driver_t *driver, jack_engine_t *engine)
 	int port_flags;
 	channel_t chn;
 	char buf[JACK_PORT_NAME_SIZE];
+	jack_latency_range_t range;
 		
 	driver->engine = engine;
 	
-	driver->engine->set_buffer_size (engine, driver->frames_per_cycle);
+	if (driver->engine->set_buffer_size (engine, driver->frames_per_cycle)) {
+		jack_error ("portaudio: cannot set engine buffer size to %d (check MIDI)", driver->frames_per_cycle);
+		return -1;
+	}
 	driver->engine->set_sample_rate (engine, driver->frame_rate);
 	
 	port_flags = JackPortIsOutput|JackPortIsPhysical|JackPortIsTerminal;
@@ -158,7 +162,8 @@ portaudio_driver_attach (portaudio_driver_t *driver, jack_engine_t *engine)
 			/* XXX fix this so that it can handle: systemic (external) latency
 			*/
 
-			jack_port_set_latency (port, driver->frames_per_cycle);
+			range.min = range.max = driver->frames_per_cycle;
+			jack_port_set_latency_range (port, JackCaptureLatency, &range);
 			driver->capture_ports = jack_slist_append (driver->capture_ports, port);
 	}
 	
@@ -176,7 +181,8 @@ portaudio_driver_attach (portaudio_driver_t *driver, jack_engine_t *engine)
 			/* XXX fix this so that it can handle: systemic (external) latency
 			*/
 	
-			jack_port_set_latency (port, driver->frames_per_cycle);
+			range.min = range.max = driver->frames_per_cycle;
+			jack_port_set_latency_range (port, JackPlaybackLatency, &range);
 			driver->playback_ports = jack_slist_append (driver->playback_ports, port);
 	}
 
@@ -241,7 +247,7 @@ portaudio_driver_read (portaudio_driver_t *driver, jack_nframes_t nframes)
 	}
    
 	driver->engine->transport_cycle_start (driver->engine,
-					   jack_get_microseconds ());
+					   driver->engine->get_microseconds ());
 	return 0;
 }          
 
@@ -323,8 +329,12 @@ portaudio_driver_set_parameters (portaudio_driver_t* driver,
 
 		/* tell engine about buffer size */
 		if (driver->engine) {
-			driver->engine->set_buffer_size (
-				driver->engine, driver->frames_per_cycle);
+			if (driver->engine->set_buffer_size (
+				    driver->engine, driver->frames_per_cycle)) {
+				jack_error ("portaudio: cannot set engine buffer size to %d (check MIDI)", driver->frames_per_cycle);
+				return -1;
+			}
+				
 		}
 		return 0;
 
@@ -332,9 +342,9 @@ portaudio_driver_set_parameters (portaudio_driver_t* driver,
 
 		// JOQ: this driver is dead.  How do we terminate it?
 		Pa_Terminate();
-		fprintf(stderr, "Unable to set portaudio parameters\n"); 
-		fprintf(stderr, "Error number: %d\n", err);
-		fprintf(stderr, "Error message: %s\n", Pa_GetErrorText(err));
+		jack_error("Unable to set portaudio parameters");
+		jack_error("Error number: %d", err);
+		jack_error("Error message: %s", Pa_GetErrorText(err));
 		return EIO;
 	}
 }
@@ -345,8 +355,8 @@ portaudio_driver_reset_parameters (portaudio_driver_t* driver,
 				   jack_nframes_t rate)
 {
 	if (!jack_power_of_two(nframes)) {
-		fprintf (stderr, "PA: frames must be a power of two "
-			 "(64, 512, 1024, ...)\n");
+		jack_error("PA: frames must be a power of two "
+			 "(64, 512, 1024, ...)");
 		return EINVAL;
 	}
 
@@ -541,7 +551,7 @@ portaudio_driver_new (char *name,
 	jack_driver_init ((jack_driver_t *) driver);
 
 	if (!jack_power_of_two(frames_per_cycle)) {
-		fprintf (stderr, "PA: -p must be a power of two.\n");
+		jack_error ("PA: -p must be a power of two.");
 		goto error;
 	}
 
@@ -653,9 +663,9 @@ portaudio_driver_new (char *name,
 error:
 
 	Pa_Terminate();
-	fprintf(stderr, "An error occured while using the portaudio stream\n"); 
-	fprintf(stderr, "Error number: %d\n", err);
-	fprintf(stderr, "Error message: %s\n", Pa_GetErrorText(err));
+	jack_error("An error occured while using the portaudio stream");
+	jack_error("Error number: %d", err);
+	jack_error("Error message: %s", Pa_GetErrorText(err));
 	free(driver);
 	return NULL;
 }
