@@ -51,7 +51,13 @@ typedef struct _jack_port_buffer_list {
     jack_port_buffer_info_t *info;	/* jack_buffer_info_t array */
 } jack_port_buffer_list_t;
 
+typedef struct _jack_reserved_name {
+    jack_client_id_t uuid;
+    char name[JACK_CLIENT_NAME_SIZE];
+} jack_reserved_name_t;
+
 #define JACKD_WATCHDOG_TIMEOUT 10000
+#define JACKD_CLIENT_EVENT_TIMEOUT 2000
 
 /* The main engine structure in local memory. */
 struct _jack_engine {
@@ -62,6 +68,8 @@ struct _jack_engine {
     jack_driver_desc_t    *driver_desc;
     JSList                *driver_params;
 
+    JSList                *slave_drivers;
+
     /* these are "callbacks" made by the driver backend */
     int  (*set_buffer_size) (struct _jack_engine *, jack_nframes_t frames);
     int  (*set_sample_rate) (struct _jack_engine *, jack_nframes_t frames);
@@ -70,13 +78,14 @@ struct _jack_engine {
     void (*delay)	    (struct _jack_engine *, float delayed_usecs);
     void (*transport_cycle_start) (struct _jack_engine *, jack_time_t time);
     void (*driver_exit)     (struct _jack_engine *);
-
+    jack_time_t (*get_microseconds)(void);
     /* "private" sections starts here */
 
     /* engine serialization -- use precedence for deadlock avoidance */
-    pthread_mutex_t request_lock;	/* precedes client_lock */
-    pthread_mutex_t client_lock;
+    pthread_mutex_t request_lock; /* precedes client_lock */
+    pthread_rwlock_t client_lock;
     pthread_mutex_t port_lock;
+    pthread_mutex_t problem_lock; /* must hold write lock on client_lock */
     int		    process_errors;
     int		    period_msecs;
 
@@ -99,6 +108,7 @@ struct _jack_engine {
     pthread_t	    watchdog_thread;
 
     int		    fds[2];
+    int		    cleanup_fifo[2];
     jack_client_id_t next_client_id;
     size_t	    pfd_size;
     size_t	    pfd_max;
@@ -106,9 +116,17 @@ struct _jack_engine {
     char	    fifo_prefix[PATH_MAX+1];
     int		   *fifo;
     unsigned long   fifo_size;
+
+    /* session handling */
+    int		    session_reply_fd;
+    int		    session_pending_replies;
+
     unsigned long   external_client_cnt;
     int		    rtpriority;
-    char	    freewheeling;
+    volatile char   freewheeling;
+    volatile char   stop_freewheeling;
+    jack_client_id_t fwclient;
+    pthread_t       freewheel_thread;
     char	    verbose;
     char	    do_munlock;
     const char	   *server_name;
@@ -116,13 +134,18 @@ struct _jack_engine {
     int		    reordered;
     int		    watchdog_check;
     int		    feedbackcount;
+    int             removing_clients;
     pid_t           wait_pid;
-    pthread_t       freewheel_thread;
     int             nozombies;
+    int             timeout_count_threshold;
+    volatile int    problems;
+    volatile int    timeout_count;
+    volatile int    new_clients_allowed;    
 
     /* these lists are protected by `client_lock' */
     JSList	   *clients;
     JSList	   *clients_waiting;
+    JSList	   *reserved_client_names;
 
     jack_port_internal_t    *internal_ports;
     jack_client_internal_t  *timebase_client;
@@ -162,6 +185,7 @@ jack_engine_t  *jack_engine_new (int real_time, int real_time_priority,
 				 int verbose, int client_timeout,
 				 unsigned int port_max,
                                  pid_t waitpid, jack_nframes_t frame_time_offset, int nozombies, 
+				 int timeout_count_threshold,
 				 JSList *drivers);
 void		jack_engine_delete (jack_engine_t *);
 int		jack_run (jack_engine_t *engine);
@@ -169,6 +193,9 @@ int		jack_wait (jack_engine_t *engine);
 int		jack_engine_load_driver (jack_engine_t *engine,
 					 jack_driver_desc_t * driver_desc,
 					 JSList * driver_params);
+int		jack_engine_load_slave_driver (jack_engine_t *engine,
+					       jack_driver_desc_t * driver_desc,
+					       JSList * driver_params);
 void		jack_dump_configuration(jack_engine_t *engine, int take_lock);
 
 /* private engine functions */
@@ -182,22 +209,38 @@ extern jack_timer_type_t clock_source;
 extern jack_client_internal_t *
 jack_client_internal_by_id (jack_engine_t *engine, jack_client_id_t id);
 
-static inline void jack_lock_graph (jack_engine_t* engine) {
-	DEBUG ("acquiring graph lock");
-	pthread_mutex_lock (&engine->client_lock);
+#define jack_rdlock_graph(e) { DEBUG ("acquiring graph read lock"); if (pthread_rwlock_rdlock (&e->client_lock)) abort(); }
+#define jack_lock_graph(e) { DEBUG ("acquiring graph write lock"); if (pthread_rwlock_wrlock (&e->client_lock)) abort(); }
+#define jack_try_rdlock_graph(e) pthread_rwlock_tryrdlock (&e->client_lock)
+#define jack_unlock_graph(e) { DEBUG ("release graph lock"); if (pthread_rwlock_unlock (&e->client_lock)) abort(); }
+
+#define jack_trylock_problems(e) pthread_mutex_trylock (&e->problem_lock)
+#define jack_lock_problems(e) { DEBUG ("acquiring problem lock"); if (pthread_mutex_lock (&e->problem_lock)) abort(); }
+#define jack_unlock_problems(e) { DEBUG ("release problem lock"); if (pthread_mutex_unlock (&e->problem_lock)) abort(); }
+
+#if 0
+static inline void jack_rdlock_graph (jack_engine_t* engine) {
+	DEBUG ("acquiring graph read lock");
+	pthread_rwlock_rdlock (&engine->client_lock);
 }
 
-static inline int jack_try_lock_graph (jack_engine_t *engine)
+static inline void jack_lock_graph (jack_engine_t* engine) {
+	DEBUG ("acquiring graph lock");
+	pthread_rwlock_wrlock (&engine->client_lock);
+}
+
+static inline int jack_try_rdlock_graph (jack_engine_t *engine)
 {
-	DEBUG ("TRYING to acquiring graph lock");
-	return pthread_mutex_trylock (&engine->client_lock);
+	DEBUG ("TRYING to acquiring graph read lock");
+	return pthread_rwlock_tryrdlock (&engine->client_lock);
 }
 
 static inline void jack_unlock_graph (jack_engine_t* engine) 
 {
 	DEBUG ("releasing graph lock");
-	pthread_mutex_unlock (&engine->client_lock);
+	pthread_rwlock_unlock (&engine->client_lock);
 }
+#endif
 
 static inline unsigned int jack_power_of_two (unsigned int n)
 {
@@ -210,5 +253,20 @@ void	jack_port_clear_connections (jack_engine_t *engine,
 void	jack_port_registration_notify (jack_engine_t *, jack_port_id_t, int);
 void	jack_port_release (jack_engine_t *engine, jack_port_internal_t *);
 void	jack_sort_graph (jack_engine_t *engine);
+int     jack_stop_freewheeling (jack_engine_t* engine, int engine_exiting);
+jack_client_internal_t *
+jack_client_by_name (jack_engine_t *engine, const char *name);
+
+int  jack_deliver_event (jack_engine_t *, jack_client_internal_t *, jack_event_t *);
+void jack_stop_watchdog (jack_engine_t * );
+
+void
+jack_engine_signal_problems (jack_engine_t* engine);
+int
+jack_use_driver (jack_engine_t *engine, struct _jack_driver *driver);
+int
+jack_drivers_start (jack_engine_t *engine);
+int
+jack_add_slave_driver (jack_engine_t *engine, struct _jack_driver *driver);
 
 #endif /* __jack_engine_h__ */
