@@ -39,6 +39,7 @@
 #include <jack/shm.h>
 #include <jack/driver_parse.h>
 #include <jack/messagebuffer.h>
+#include <jack/midiport.h>
 
 #ifdef USE_CAPABILITIES
 
@@ -56,7 +57,7 @@ static JSList *drivers = NULL;
 static sigset_t signals;
 static jack_engine_t *engine = NULL;
 static char *server_name = NULL;
-static int realtime = 0;
+static int realtime = 1;
 static int realtime_priority = 10;
 static int do_mlock = 1;
 static int temporary = 0;
@@ -66,6 +67,12 @@ static unsigned int port_max = 256;
 static int do_unlock = 0;
 static jack_nframes_t frame_time_offset = 0;
 static int nozombies = 0;
+static int timeout_count_threshold = 0;
+
+extern int sanitycheck (int, int);
+
+static jack_driver_desc_t *
+jack_find_driver_descriptor (const char * name);
 
 static void 
 do_nothing_handler (int sig)
@@ -81,13 +88,14 @@ do_nothing_handler (int sig)
 }
 
 static int
-jack_main (jack_driver_desc_t * driver_desc, JSList * driver_params)
+jack_main (jack_driver_desc_t * driver_desc, JSList * driver_params, JSList * slave_names)
 {
 	int sig;
 	int i;
 	sigset_t allsignals;
 	struct sigaction action;
 	int waiting;
+	JSList * node;
 
 	/* ensure that we are in our own process group so that
 	   kill (SIG, -pgrp) does the right thing.
@@ -147,7 +155,7 @@ jack_main (jack_driver_desc_t * driver_desc, JSList * driver_params)
 				       do_mlock, do_unlock, server_name,
 				       temporary, verbose, client_timeout,
 				       port_max, getpid(), frame_time_offset, 
-				       nozombies, drivers)) == 0) {
+				       nozombies, timeout_count_threshold, drivers)) == 0) {
 		jack_error ("cannot create engine");
 		return -1;
 	}
@@ -160,7 +168,16 @@ jack_main (jack_driver_desc_t * driver_desc, JSList * driver_params)
 		goto error;
 	}
 
-	if (engine->driver->start (engine->driver) != 0) {
+	for (node=slave_names; node; node=jack_slist_next(node)) {
+		char *sl_name = node->data;
+		jack_driver_desc_t *sl_desc = jack_find_driver_descriptor(sl_name);
+		if (sl_desc) {
+			jack_engine_load_slave_driver(engine, sl_desc, NULL);
+		}
+	}
+
+
+	if (jack_drivers_start (engine) != 0) {
 		jack_error ("cannot start driver");
 		goto error;
 	}
@@ -352,7 +369,7 @@ jack_drivers_load ()
 static void copyright (FILE* file)
 {
 	fprintf (file, "jackd " VERSION "\n"
-"Copyright 2001-2005 Paul Davis and others.\n"
+"Copyright 2001-2009 Paul Davis, Stephane Letz, Jack O'Quinn, Torben Hohn and others.\n"
 "jackd comes with ABSOLUTELY NO WARRANTY\n"
 "This is free software, and you are welcome to redistribute it\n"
 "under certain conditions; see the file COPYING for details\n\n");
@@ -362,22 +379,28 @@ static void usage (FILE *file)
 {
 	copyright (file);
 	fprintf (file, "\n"
-"usage: jackd [ --realtime OR -R [ --realtime-priority OR -P priority ] ]\n"
+"usage: jackd [ --no-realtime OR -r ]\n"
+"             [ --realtime OR -R [ --realtime-priority OR -P priority ] ]\n"
+"      (the two previous arguments are mutually exclusive. The default is --realtime)\n"
 "             [ --name OR -n server-name ]\n"
 "             [ --no-mlock OR -m ]\n"
 "             [ --unlock OR -u ]\n"
 "             [ --timeout OR -t client-timeout-in-msecs ]\n"
 "             [ --port-max OR -p maximum-number-of-ports]\n"
 "             [ --debug-timer OR -D ]\n"
+"             [ --no-sanity-checks OR -N ]\n"
 "             [ --verbose OR -v ]\n"
 "             [ --clocksource OR -c [ c(ycle) | h(pet) | s(ystem) ]\n"
-"             [ --replace-registry OR -r ]\n"
+"             [ --replace-registry ]\n"
 "             [ --silent OR -s ]\n"
 "             [ --version OR -V ]\n"
 "             [ --nozombies OR -Z ]\n"
 "         -d backend [ ... backend args ... ]\n"
-"             The backend can be `alsa', `coreaudio', `dummy',\n"
-"                                `freebob', `oss', `sun', or `portaudio'.\n\n"
+#ifdef __APPLE__
+"             Available backends may include: coreaudio, dummy, net, portaudio.\n\n"
+#else 
+"             Available backends may include: alsa, dummy, freebob, firewire, net, oss, sun, or portaudio.\n\n"
+#endif
 "       jackd -d backend --help\n"
 "             to display options for each backend\n\n");
 }	
@@ -509,26 +532,38 @@ main (int argc, char *argv[])
 
 {
 	jack_driver_desc_t * desc;
-	const char *options = "-ad:P:uvshVRZTFlt:mn:p:c:";
+	int replace_registry = 0;
+	int do_sanity_checks = 1;
+	int show_version = 0;
+
+	const char *options = "-d:P:uvshVrRZTFlt:mM:n:Np:c:X:C:";
 	struct option long_options[] = 
 	{ 
+		/* keep ordered by single-letter option code */
+
+		{ "clock-source", 1, 0, 'c' },
 		{ "driver", 1, 0, 'd' },
-		{ "verbose", 0, 0, 'v' },
 		{ "help", 0, 0, 'h' },
 		{ "tmpdir-location", 0, 0, 'l' },
-		{ "port-max", 1, 0, 'p' },
 		{ "no-mlock", 0, 0, 'm' },
+		{ "midi-bufsize", 1, 0, 'M' },
 		{ "name", 1, 0, 'n' },
-		{ "unlock", 0, 0, 'u' },
-		{ "realtime", 0, 0, 'R' },
-		{ "replace-registry", 0, 0, 'r' },
+                { "no-sanity-checks", 0, 0, 'N' },
+		{ "port-max", 1, 0, 'p' },
 		{ "realtime-priority", 1, 0, 'P' },
+		{ "no-realtime", 0, 0, 'r' },
+		{ "realtime", 0, 0, 'R' },
+		{ "replace-registry", 0, &replace_registry, 0 },
+		{ "silent", 0, 0, 's' },
+		{ "sync", 0, 0, 'S' },
 		{ "timeout", 1, 0, 't' },
 		{ "temporary", 0, 0, 'T' },
+		{ "unlock", 0, 0, 'u' },
 		{ "version", 0, 0, 'V' },
-		{ "silent", 0, 0, 's' },
-		{ "clock-source", 1, 0, 'c' },
+		{ "verbose", 0, 0, 'v' },
+		{ "slave-driver", 1, 0, 'X' },
 		{ "nozombies", 0, 0, 'Z' },
+		{ "timeout-thres", 2, 0, 'C' },
 		{ 0, 0, 0, 0 }
 	};
 	int opt = 0;
@@ -537,9 +572,9 @@ main (int argc, char *argv[])
 	char *driver_name = NULL;
 	char **driver_args = NULL;
 	JSList * driver_params;
+	JSList * slave_drivers = NULL;
+	size_t midi_buffer_size = 0;
 	int driver_nargs = 1;
-	int show_version = 0;
-	int replace_registry = 0;
 	int i;
 	int rc;
 
@@ -566,6 +601,13 @@ main (int argc, char *argv[])
 			}
 			break;
 
+		case 'C':
+			if (optarg)
+				timeout_count_threshold = atoi (optarg);
+			else
+				timeout_count_threshold = 250;
+			break;
+
 		case 'd':
 			seen_driver = 1;
 			driver_name = optarg;
@@ -584,10 +626,18 @@ main (int argc, char *argv[])
 			do_mlock = 0;
 			break;
 
+		case 'M':
+			midi_buffer_size = (unsigned int) atol (optarg);
+			break;
+
 		case 'n':
 			server_name = optarg;
 			break;
 
+		case 'N':
+			do_sanity_checks = 0;
+			break;
+
 		case 'p':
 			port_max = (unsigned int) atol (optarg);
 			break;
@@ -597,10 +647,11 @@ main (int argc, char *argv[])
 			break;
 
 		case 'r':
-			replace_registry = 1;
+			realtime = 0;
 			break;
 
 		case 'R':
+			/* this is now the default */
 			realtime = 1;
 			break;
 
@@ -608,6 +659,10 @@ main (int argc, char *argv[])
 			jack_set_error_function (silent_jack_error_callback);
 			break;
 
+                case 'S':
+                        /* this option is for jack2 only (synchronous mode) */
+                        break;
+
 		case 'T':
 			temporary = 1;
 			break;
@@ -628,6 +683,9 @@ main (int argc, char *argv[])
 			show_version = 1;
 			break;
 
+		case 'X':
+			slave_drivers = jack_slist_append(slave_drivers, optarg);
+			break;
 		case 'Z':
 			nozombies = 1;
 			break;
@@ -647,6 +705,12 @@ main (int argc, char *argv[])
 				" tmpdir " DEFAULT_TMP_DIR 
 				" protocol " PROTOCOL_VERSION
 				"\n");
+		return 0;
+	}
+
+	copyright (stdout);
+
+	if (do_sanity_checks && (0 < sanitycheck (realtime, (clock_source == JACK_TIMER_CYCLE_COUNTER)))) {
 		return -1;
 	}
 
@@ -666,6 +730,15 @@ main (int argc, char *argv[])
 		fprintf (stderr, "jackd: no drivers found; exiting\n");
 		exit (1);
 	}
+	
+	if (midi_buffer_size != 0) {
+		jack_port_type_info_t* port_type = &jack_builtin_port_types[JACK_MIDI_PORT_TYPE];
+		port_type->buffer_size = midi_buffer_size * jack_midi_internal_event_size ();
+		port_type->buffer_scale_factor = -1;
+		if (verbose) {
+			fprintf (stderr, "Set MIDI buffer size to %u bytes\n", port_type->buffer_size);
+		}
+	}
 
 	desc = jack_find_driver_descriptor (driver_name);
 	if (!desc) {
@@ -700,8 +773,6 @@ main (int argc, char *argv[])
 	if (server_name == NULL)
 		server_name = jack_default_server_name ();
 
-	copyright (stdout);
-
 	rc = jack_register_server (server_name, replace_registry);
 	switch (rc) {
 	case EEXIST:
@@ -725,7 +796,7 @@ main (int argc, char *argv[])
 	jack_cleanup_files (server_name);
 
 	/* run the server engine until it terminates */
-	jack_main (desc, driver_params);
+	jack_main (desc, driver_params, slave_drivers);
 
 	/* clean up shared memory and files from this server instance */
 	if (verbose)
